"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.runWithMutexW = exports.runWithMutexR = exports.RWMutex = void 0;
const events_1 = require("events");
/**Refer to RWMutex in Golang */
class RWMutex extends events_1.EventEmitter {
    constructor() {
        super();
        this._rlocks = 0;
        this._wlocked = false;
        this._lockQueue = [];
        this.state = 'UNLOCKED';
        this.setMaxListeners(1000);
        this.on('unlock', () => this._releaseLocks());
    }
    _releaseLocks() {
        const length = this._lockQueue.length;
        for (let i = 0; i < length; i++) {
            const lock = this._lockQueue.shift();
            if (lock.type === 'R') {
                lock.res();
                continue;
            }
            if (i === 0) {
                lock.res();
            }
            else {
                this._lockQueue.unshift(lock);
            }
            break;
        }
    }
    _enqueueLockAndWait(type) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((res) => {
                this._lockQueue.push({ res, type });
            });
        });
    }
    writeLength() {
        return this._lockQueue.filter((q) => q.type === 'W').length;
    }
    hasNextWrites() {
        return this._lockQueue.filter((q) => q.type === 'W').length > 0;
    }
    hasNextRead() {
        return this._lockQueue.filter((q) => q.type === 'R').length > 0;
    }
    /**
     * @deprecated use .rLock()
     */
    RLock() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.rLock();
        });
    }
    /**Acquire Read Lock. Use it for thread-safe operations */
    rLock() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this._wlocked || this._lockQueue.some((lock) => lock.type === 'W')) {
                yield this._enqueueLockAndWait('R');
            }
            this._rlocks++;
            this.state = 'RLOCKED';
        });
    }
    /**
     * @deprecated use .lock()
     */
    Lock() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.lock();
        });
    }
    /**Acquire Write Lock. Use it for operations that require monopoly resource usage */
    lock() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this._rlocks > 0 || this._wlocked) {
                yield this._enqueueLockAndWait('W');
            }
            this._wlocked = true;
            this.state = 'LOCKED';
        });
    }
    /**
     * @deprecated use .rUnlock()
     */
    RUnlock() {
        return this.rUnlock();
    }
    /**Release Read Lock */
    rUnlock() {
        if (this._rlocks < 1) {
            throw new Error('Cannot RUnlock not rlocked mutex. Fix race conditions');
        }
        this._rlocks--;
        if (this._rlocks === 0) {
            this.state = 'UNLOCKED';
            this.emit('unlock');
        }
    }
    /**
     * @deprecated use .unlock()
     */
    Unlock() {
        return this.unlock();
    }
    /**Release Write Lock */
    unlock() {
        if (!this._wlocked) {
            throw new Error(`Cannot Unlock unlocked mutex. Fix race conditions`);
        }
        this._wlocked = false;
        this.state = 'UNLOCKED';
        setImmediate(() => {
            this.emit('unlock');
        });
    }
}
exports.RWMutex = RWMutex;
function runWithMutexR(mx, fnc) {
    return __awaiter(this, void 0, void 0, function* () {
        yield mx.rLock();
        try {
            const res = yield fnc();
            try {
                mx.rUnlock();
            }
            catch (err) {
                throw new Error(`Cannot RUnlock after the promise was resolved. Error: ${err}`);
            }
            return res;
        }
        catch (err) {
            try {
                mx.rUnlock();
            }
            catch (err1) {
                throw new Error(`Cannot RUnlock after the promise was rejected with error: ${err}. Unlock error: ${err1}`);
            }
            throw err;
        }
    });
}
exports.runWithMutexR = runWithMutexR;
function runWithMutexW(mx, fnc) {
    return __awaiter(this, void 0, void 0, function* () {
        yield mx.lock();
        try {
            const res = yield fnc();
            try {
                mx.unlock();
            }
            catch (err) {
                throw new Error(`Cannot Unlock after the promise was resolved. Error: ${err}`);
            }
            return res;
        }
        catch (err) {
            try {
                mx.unlock();
            }
            catch (err1) {
                throw new Error(`Cannot unlock after the promise was rejected with error: ${err}. Unlock error: ${err1}`);
            }
            throw err;
        }
    });
}
exports.runWithMutexW = runWithMutexW;
